unit UPascalParser;

{$mode delphi}

interface

uses
  Classes, SysUtils, UParser, USourceCode, Dialogs;

type
  TGetSourceEvent = function (Name: string; Source: TStringList): Boolean of object;

  { TPascalParser }

  TPascalParser = class(TBaseParser)
  private
    FOnGetSource: TGetSourceEvent;
  public
    function ParseFile(Name: string): Boolean;
    function ParseWhileDo(var WhileDo: TWhileDo; SourceCode: TCommonBlock): Boolean;
    function ParseExpression(SourceCode: TExpression): Boolean;
    function ParseExpressionParenthases(SourceCode: TExpression;
      Expressions: TListExpression): Boolean;
    function ParseExpressionOperator(SourceCode: TExpression;
      Expressions: TListExpression): Boolean;
    function ParseExpressionRightValue(SourceCode: TExpression;
      Expressions: TListExpression): Boolean;
    function ParseExpressionFunctionCall(SourceCode: TExpression;
      Expressions: TListExpression; var Func: TFunctionCall): Boolean;
    function ParseUses(SourceCode: TUsedModuleList; AExported: Boolean): Boolean;
    function ParseModule(ProgramCode: TProgram): TModule;
    function ParseUnit(var SourceCode: TModuleUnit; ProgramCode: TProgram): Boolean;
    function ParseUnitInterface(SourceCode: TModuleUnit): Boolean;
    function ParseUnitImplementation(SourceCode: TModuleUnit): Boolean;
    procedure ParseProgram(SourceCode: TModuleProgram);
    procedure ParseCommonBlock(SourceCode: TCommonBlock; EndSymbol: char = ';';
      WithBody: Boolean = True);
    procedure ParseCommonBlockInterface(SourceCode: TCommonBlock);
    function ParseCommand(SourceCode: TCommonBlock): TCommand;
    function ParseBeginEnd(var BeginEnd: TBeginEnd; SourceCode: TCommonBlock): Boolean;
    function ParseFunctionList(SourceCode: TFunctionList; Exported: Boolean = False): Boolean;
    procedure ParseFunctionParameters(SourceCode: TFunction; ValidateParams: Boolean = False);
    function ParseIfThenElse(var IfThenElse: TIfThenElse; SourceCode: TCommonBlock): Boolean;
    function ParseForToDo(var ForToDo: TForToDo; SourceCode: TCommonBlock): Boolean;
    function ParseAssigment(var Assignment: TAssignment; SourceCode: TCommonBlock): Boolean;
    function ParseFunctionCall(var Call: TFunctionCall; SourceCode: TCommonBlock): Boolean;
    function ParseVariableList(SourceCode: TVariableList; Exported: Boolean = False): Boolean;
    procedure ParseVariable(SourceCode: TVariableList; Exported: Boolean = False);
    function ParseConstantList(SourceCode: TConstantList; Exported: Boolean = False): Boolean;
    function ParseConstant(SourceCode: TConstantList; Exported: Boolean = False): Boolean;
    function ParseTypeList(SourceCode: TTypeList; Exported: Boolean = False;
      AssignSymbol: string = '='): Boolean;
    function ParseType(TypeList: TTypeList; ExpectName: Boolean = True;
      AssignSymbol: string = '='; ForwardDeclaration: Boolean = False): TType;
    function ParseTypeSubType(TypeList: TTypeList; Name: string;
      ExpectName: Boolean; ForwardDeclaration: Boolean): TType;
    function ParseTypeBase(TypeList: TTypeList; Name: string): TType;
    function ParseTypePointer(TypeList: TTypeList; Name: string): TType;
    function ParseTypeEnumeration(TypeList: TTypeList; Name: string): TType;
    function ParseTypeRecord(TypeList: TTypeList; Name: string): TType;
    function ParseTypeClass(TypeList: TTypeList; Name: string): TType;
    function ParseTypeArray(TypeList: TTypeList; Name: string): TType;
    function ParseTypeSubRange(TypeList: TTypeList; Name: string): TType;
    property OnGetSource: TGetSourceEvent read FOnGetSource
      write FOnGetSource;
    constructor Create;
    destructor Destroy; override;
  end;


resourcestring
  SUnknownIdentifier = 'Unknown identificator "%s".';
  SIllegalExpression = 'Illegal expression "%s".';
  SRedefineIdentifier = 'Identificator "%s" redefinition.';
  SEndOfDataReached = 'Parser reached to end of input data.';
  SUndefinedVariable = 'Undefined variable "%s".';
  SUndefinedType = 'Undefined type "%s".';
  SUndefinedConstant = 'Undefined constant "%s".';
  SUnitNotFound = 'Unit "%s" not found.';
  SFunctionNotDeclared = 'Function "%s" not declared.';
  SUnknownProcName = 'Unknown proc name "%s".';
  SUnknownModuleType = 'Unknown module name "%s".';
  SInvalidConstruction = 'Invalid construction.';
  SInvalidAssignmentValue = 'Invalid assignment "%s".';
  SParamDiffers = 'Declaration of parametr "%s" differs.';

implementation

{ TPascalParser }

function TPascalParser.ParseFile(Name: string): Boolean;
var
  Parser: TPascalParser;
  NewModule: TModule;
begin
  try
    Parser := TPascalParser.Create;
    Parser.SourceCodeText := TStringList.Create;
    Parser.OnDebugLog := OnDebugLog;
    Parser.ProgramCode := ProgramCode;
    Parser.OnGetSource := OnGetSource;
    if Assigned(OnGetSource) then begin
      if FOnGetSource(Name, Parser.SourceCodeText) then begin
        Parser.Process;
        Parser.FileName := Name;
        Parser.OnErrorMessage := OnErrorMessage;
        //NewModule :=
        Parser.ParseModule(ProgramCode);
        //ProgramCode.Modules.Add(NewModule);
        Result := True;
      end else Result := False;
    end else Result := False;
  finally
    Parser.SourceCodeText.Free;
    Parser.Free;
  end;
end;

function TPascalParser.ParseWhileDo(var WhileDo: TWhileDo; SourceCode: TCommonBlock): Boolean;
begin
  if NextToken = 'while' then begin
    Expect('while');
    WhileDo := TWhileDo.Create;
    WhileDo.CommonBlock := SourceCode;
    with WhileDo do begin
      Condition.CommonBlock := CommonBlock;
      ParseExpression(Condition);
      Expect('do');
      Command := ParseCommand(CommonBlock);
    end;
    Result := True;
  end else Result := False;
end;

{ TExpression }

function TPascalParser.ParseExpression(SourceCode: TExpression): Boolean;
var
  Expressions: TListExpression;
  I: Integer;
  II: Integer;
begin
  try
    Expressions := TListExpression.Create;
    Expressions.OwnsObjects := False;
    Expressions.Add(TExpression.Create);
    with SourceCode do begin
      while ((NextToken <> ';') and (NextToken <> ',') and (not IsKeyWord(NextToken))) and not
        (((NextToken = ')') or (NextToken = ']'))) and not (NextTokenType = ttEndOfFile) do begin
        if not ParseExpressionParenthases(SourceCode, Expressions) then
        if not ParseExpressionOperator(SourceCode, Expressions) then
        if not ParseExpressionRightValue(SourceCode, Expressions) then
        begin
          ErrorMessage(SInvalidAssignmentValue, [NextToken]);
          ReadToken;
        end;
      end;

      // Build expression tree using operator precedence
      for II := 0 to High(Operators) do begin
        I := 1;
        while (I < Expressions.Count - 1) do begin
          if not Expressions[I].Associated and
            (Expressions[I].OperatorName = Operators[II]) then
          begin
            Expressions[I].Associated := True;
            Expressions[I - 1].SubItems.Last := Expressions[I];
            Expressions[I + 1].SubItems.First := Expressions[I];
            Expressions.Delete(I);
          end else Inc(I);
        end;
      end;
      if Assigned(Expressions.First.SubItems.Last) then
        Assign(Expressions.First.SubItems.Last);
      Expressions.First.SubItems.Last := nil;
      //ShowMessage(IntToStr(Expressions.Count));
      if Expressions.Count > 1 then
        Expressions[1].SubItems.First := nil;
    end;
  finally
    Expressions.Free;
  end;
end;

function TPascalParser.ParseExpressionParenthases(SourceCode: TExpression;
  Expressions: TListExpression): Boolean;
var
  NewExpression: TExpression;
begin
  if NextToken = '(' then begin
    Expect('(');
    // Subexpression
    NewExpression := TExpression.Create;
    NewExpression.Braces := True;
    NewExpression.CommonBlock := SourceCode.CommonBlock;
    ParseExpression(NewExpression);

    Expressions.Last.SubItems.Last := NewExpression;
    with Expressions.Items[Expressions.Add(TExpression.Create)] do
    begin
      CommonBlock := SourceCode.CommonBlock;
      SubItems.First := NewExpression;
    end;
    Expect(')');
    Result := True;
  end else Result := False;
end;

function TPascalParser.ParseExpressionOperator(SourceCode: TExpression;
  Expressions: TListExpression): Boolean;
begin
  if IsOperator(NextToken) then begin
    // Operator
    Expressions.Last.OperatorName := ReadToken;
    Expressions.Last.NodeType := ntOperator;
    Result := True;
  end else Result := False;
end;

function TPascalParser.ParseExpressionRightValue(SourceCode: TExpression;
  Expressions: TListExpression): Boolean;
var
  UseType: TType;
  UseVariable: TVariable;
  UseConstant: TConstant;
  UseFunction: TFunction;
  FunctionCall: TFunctionCall;
  NewExpression: TExpression;
  Identifier: string;
  IntConst: Integer;
begin
  NewExpression := nil;
  with SourceCode do begin
//  if IsIdentificator(NextToken) then begin
    // Start with type
    UseType := CommonBlock.Types.Search(NextToken);
    if Assigned(UseType) then begin
      ReadToken;
      if NextToken = '(' then begin
        Expect('(');
        // Typecasting
        NewExpression := TExpression.Create;
        NewExpression.CommonBlock := SourceCode.CommonBlock;
        NewExpression.NodeType := ntTypecast;
        NewExpression.UseType := UseType;
        ParseExpression(NewExpression);
        Expect(')');
      end else
      if (UseType is TTypeRecord) or (UseType is TTypeClass) then begin
        // Type context
        Expect('.');
        Identifier := ReadToken;
        UseVariable := TTypeRecord(UseType).CommonBlock.Variables.Search(Identifier);
        if Assigned(UseVariable) then begin
          // Record or class variable
          NewExpression := TExpression.Create;
          NewExpression.CommonBlock := SourceCode.CommonBlock;
          NewExpression.NodeType := ntVariable;
          NewExpression.Variable := UseVariable;
        end;
        if not Assigned(NewExpression) then begin
          UseFunction := TTypeRecord(UseType).CommonBlock.Functions.Search(Identifier);
          if Assigned(UseFunction) then begin
            // Record or class functions
            ParseExpressionFunctionCall(SourceCode, Expressions, FunctionCall);
            NewExpression := TExpression.Create;
            NewExpression.CommonBlock := SourceCode.CommonBlock;
            NewExpression.NodeType := ntFunction;
            NewExpression.FunctionCall := FunctionCall;
          end;
        end;
        if not Assigned(NewExpression) then
          ErrorMessage(SUndefinedVariable, [Identifier]);
      end else ErrorMessage(SIllegalExpression, [Identifier]);
    end;
    if not Assigned(NewExpression) then begin
      // Referenced variable
      UseVariable := CommonBlock.Variables.Search(NextToken);
      if Assigned(UseVariable) then begin
        ReadToken;
        NewExpression := TExpression.Create;
        NewExpression.CommonBlock := SourceCode.CommonBlock;
        NewExpression.NodeType := ntVariable;
        NewExpression.Variable := UseVariable;
      end;
    end;
    if not Assigned(NewExpression) then begin
      // Function call
      ParseExpressionFunctionCall(SourceCode, Expressions, FunctionCall);
      if Assigned(FunctionCall) then begin
        NewExpression := TExpression.Create;
        NewExpression.CommonBlock := SourceCode.CommonBlock;
        NewExpression.NodeType := ntFunction;
        NewExpression.FunctionCall := FunctionCall;
      end;
    end;
    if not Assigned(NewExpression) then begin
      // Referenced constant
      UseConstant := CommonBlock.Constants.Search(NextToken);
      if Assigned(UseConstant) then begin
        ReadToken;
        NewExpression := TExpression.Create;
        NewExpression.CommonBlock := SourceCode.CommonBlock;
        NewExpression.NodeType := ntConstant;
        NewExpression.Constant := UseConstant;
      end;
    end;
    if not Assigned(NewExpression) then begin
      // Constant value
      NewExpression := TExpression.Create;
      NewExpression.CommonBlock := SourceCode.CommonBlock;
      NewExpression.NodeType := ntValue;
      NewExpression.Constant := TConstant.Create;
      NewExpression.Value := ReadToken;
      if TryStrToInt(NewExpression.Value, IntConst) then
        NewExpression.Value := IntConst;
    end;
    if Assigned(NewExpression) then begin
      Expressions.Last.SubItems.Last := NewExpression;
      with Expressions.Items[Expressions.Add(TExpression.Create)] do
      begin
        CommonBlock := SourceCode.CommonBlock;
        SubItems.First := NewExpression;
      end;
      Result := True;
    end else begin
      Result := False;
      ErrorMessage(SUnknownIdentifier, [ReadToken]);
    end;
//  end else Result := False;
  end;
end;

function TPascalParser.ParseExpressionFunctionCall(SourceCode: TExpression;
  Expressions: TListExpression; var Func: TFunctionCall): Boolean;
var
  UseFunction: TFunction;
  NewExpression: TExpression;
  I: Integer;
begin
  Func := nil;
  with SourceCode do begin
    UseFunction := CommonBlock.Functions.Search(NextToken);
    if Assigned(UseFunction) then begin
      ReadToken;
      Func := TFunctionCall.Create;
      Func.CommonBlock := SourceCode.CommonBlock;
      Func.FunctionRef := UseFunction;
      if NextToken = '(' then begin
        Expect('(');
        for I := 0 to Func.FunctionRef.Parameters.Count - 1 do begin
          if I > 0 then Expect(',');
          NewExpression := TExpression.Create;
          NewExpression.CommonBlock := SourceCode.CommonBlock;
          ParseExpression(NewExpression);
          Func.ParameterExpression.Add(NewExpression);
        end;
        Expect(')');
      end;
      Result := True;
    end else Result := False;
  end;
end;

function TPascalParser.ParseCommand(SourceCode: TCommonBlock): TCommand;
var
  Identifier: string;
  Variable: TVariable;
  Method: TFunction;
  First: TOperation;
  Second: TOperation;
  StartIndex: integer;
  LoopVariable: TVariable;
  IdentName: string;
  FunctionName: string;
begin
  begin
    if not ParseBeginEnd(TBeginEnd(Result), SourceCode) then
    if not ParseIfThenElse(TIfThenElse(Result), SourceCode) then
    if not ParseWhileDo(TWhileDo(Result), SourceCode) then
    if not ParseForToDo(TForToDo(Result), SourceCode) then
    if not ParseAssigment(TAssignment(Result), SourceCode) then
    if not ParseFunctionCall(TFunctionCall(Result), SourceCode) then
    if NextToken = ';' then
      Result := nil
    else begin
      Result := nil;
      ErrorMessage(SIllegalExpression, [ReadToken], -1);
    end;
  end;
end;

{ TParserModule }

function TPascalParser.ParseModule(ProgramCode: TProgram): TModule;
begin
  Self.ProgramCode := ProgramCode;
  if not ParseUnit(TModuleUnit(Result), ProgramCode) then begin
    Result := TModuleProgram.Create;
    Result.ParentProgram := ProgramCode;
    ParseProgram(TModuleProgram(Result));
  end else
    ErrorMessage(SUnknownModuleType, [NextToken]);
end;

procedure TPascalParser.ParseProgram(SourceCode: TModuleProgram);
var
  Identifier: string;
begin
  with SourceCode do begin
    if NextToken = 'program' then begin
      Expect('program');
      Name := ReadToken;
      Expect(';');
    end else Name := '';

    // Uses section
    if NextToken = 'uses' then
      ParseUses(UsedModules, False);

    ParseCommonBlock(Body, '.');
    SourceCode.ParentProgram.Modules.Add(SourceCode);
  end;
end;

function TPascalParser.ParseUnit(var SourceCode: TModuleUnit; ProgramCode: TProgram): Boolean;
var
  NewModule: TModule;
begin
  if NextToken = 'unit' then begin
    SourceCode := TModuleUnit.Create;
    SourceCode.ParentProgram := ProgramCode;
    Expect('unit');
    with Sourcecode do begin
      Name := ReadToken;
    end;
    Expect(';');

    if not ParseUnitInterface(SourceCode) then
      ErrorMessage(SExpectedButFound, ['interface', NextToken]);

    if not ParseUnitImplementation(SourceCode) then
      ErrorMessage(SExpectedButFound, ['implementation', NextToken]);

    SourceCode.ParentProgram.Modules.Add(SourceCode);

    if NextToken = 'initialization' then begin
      Expect('initialization');
    end;
    if NextToken = 'finalization' then begin
      Expect('finalization');
    end;
    Result := True;
  end else Result := False;
end;

function TPascalParser.ParseUnitInterface(SourceCode: TModuleUnit): Boolean;
begin
  if NextToken = 'interface' then begin
    Expect('interface');
    // Uses section
    ParseUses(SourceCode.UsedModules, True);

    ParseCommonBlockInterface(SourceCode.Body);
    Result := True;
  end else Result := False;
end;

function TPascalParser.ParseUnitImplementation(SourceCode: TModuleUnit): Boolean;
begin
  if NextToken = 'implementation' then begin
    Expect('implementation');

    // Uses section
    if NextToken = 'uses' then
      ParseUses(SourceCode.UsedModules, False);

    ParseCommonBlock(SourceCode.Body, '.', False);
    Result := True;
  end else Result := False;
end;

{ TParserCommonBlock }

procedure TPascalParser.ParseCommonBlock(SourceCode: TCommonBlock;
  EndSymbol: char = ';'; WithBody: Boolean = True);
begin
  with SourceCode do begin
    while (NextToken <> EndSymbol) do begin
      if not ParseVariableList(Variables) then
      if not ParseConstantList(Constants) then
      if not ParseTypeList(Types) then
      if not ParseFunctionList(Functions) then begin
        if WithBody then
          if not ParseBeginEnd(Code, SourceCode) then
            ErrorMessage(SExpectedButFound, ['begin', NextToken]);
        Break;
      end;
    end;
    if WithBody then Expect(EndSymbol);
  end;
end;

procedure TPascalParser.ParseCommonBlockInterface(SourceCode: TCommonBlock);
begin
  with SourceCode do begin
    while (NextToken <> 'implementation') and (NextTokenType <> ttEndOfFile) do begin
      if not ParseVariableList(Variables, True) then
      if not ParseConstantList(Constants, True) then
      if not ParseTypeList(Types, True) then
      if not ParseFunctionList(Functions, True) then begin
        ErrorMessage(SUnknownIdentifier, [NextToken], -1);
        ReadToken;
      end;
    end;
  end;
end;

{ TParserBeginEnd }

function TPascalParser.ParseBeginEnd(var BeginEnd: TBeginEnd; SourceCode: TCommonBlock): Boolean;
var
  NewCommand: TCommand;
begin
  if NextToken = 'begin' then begin
    //ShowMessage(IntToStr(Integer(SourceCode)) + ' ' + IntToStr(Integer(SourceCode.CommonBlock)));
    BeginEnd := TBeginEnd.Create;
    TBeginEnd(BeginEnd).CommonBlock := SourceCode;
    with BeginEnd do
    begin
      Expect('begin');
      while (NextToken <> 'end') and (NextTokenType <> ttEndOfFile) do
      begin
        NewCommand := ParseCommand(CommonBlock);
        if Assigned(NewCommand) then
          Commands.Add(NewCommand);
       //ShowMessage(NextCode);
        if NextToken = ';' then
          ReadToken;
      end;
      Expect('end');
    end;
    Result := True;
  end else Result := False;
end;

{ TParserParseFunctionList }

function TPascalParser.ParseFunctionList(SourceCode: TFunctionList;
  Exported: Boolean = False): Boolean;
var
  NewValueType: TType;
  TypeName: string;
  UseName: string;
  I: Integer;
  UseType: TType;
  UseFunction: TFunction;
  FunctionType: TFunctionType;
  ValidParams: Boolean;
begin
  if (NextToken = 'procedure') or (NextToken = 'function') then begin
  with SourceCode do begin
    if NextToken = 'procedure' then begin
      Expect('procedure');
      FunctionType := ftProcedure;
    end else
    if NextToken = 'function' then begin
      Expect('function');
      FunctionType := ftFunction;
    end else
    if NextToken = 'constructor' then begin
      Expect('constructor');
      FunctionType := ftConstructor;
    end else
    if NextToken = 'destructor' then begin
      Expect('destructor');
      FunctionType := ftDestructor;
    end else ErrorMessage(SUnknownProcName, [NextToken]);

    // Read function name
    UseName := ReadToken;
    UseType := SourceCode.Parent.Types.Search(UseName);
    if Assigned(UseType) and ((UseType is TTypeRecord) or
    (UseType is TTypeClass)) then begin
      Expect('.');
      ValidParams := True;
      UseName := ReadToken;
      if UseType is TTypeRecord then begin
        UseFunction := TTypeRecord(UseType).CommonBlock.Functions.Search(UseName);
        if not Assigned(UseFunction) then begin
          ErrorMessage(SFunctionNotDeclared, [UseName]);
          Exit;
        end;
      end;
    end else begin
      // Create new function
      UseFunction := TFunction.Create;
      UseFunction.Parent := SourceCode.Parent;
      UseFunction.Name := UseName;
      UseFunction.FunctionType := FunctionType;
      Add(UseFunction);
      ValidParams := False;
    end;
    with UseFunction do begin
      // Parse parameters
      if NextToken = '(' then
        ParseFunctionParameters(UseFunction, ValidParams);

      // Parse function result type
      if FunctionType = ftFunction then begin
        Expect(':');
        TypeName := ReadToken;
        NewValueType := Parent.Types.Search(TypeName);
        if not Assigned(NewValueType) then
          ErrorMessage(SUndefinedType, [TypeName], -1);
(*        else
          begin
            ResultType := NewValueType;
            with TVariable(Parent.Variables.Items[Parent.Variables.Add(
                TVariable.Create)]) do
            begin
              Name := 'Result';
              ValueType := NewValueType;
            end;
          end;  *)
      end;
      Expect(';');

      // Check directives
      if NextToken = 'internal' then begin
        Expect('internal');
        Expect(';');
        Internal := True;
      end;
    end;

    if not Exported then ParseCommonBlock(UseFunction);
//    if UseFunction then UseFunction.Code ;
  end;
    Result := True;
  end else Result := False;
end;

procedure TPascalParser.ParseFunctionParameters(SourceCode: TFunction;
  ValidateParams: Boolean = False);
var
  Identifiers: TStringList;
  VariableName: string;
  UseVariable: TParameter;
  TypeName: string;
  UseType: TType;
  I: Integer;
begin
  with SourceCode do
  try
    Identifiers := TStringList.Create;
        Expect('(');
        while (NextToken <> ')') and (NextTokenType <> ttEndOfFile) do begin
          // while IsIdentificator(NextCode) do begin
          with TParameterList(Parameters) do begin
            VariableName := ReadToken;
            if VariableName = 'var' then begin
            end else
            if VariableName = 'const' then begin
            end else begin
              UseVariable := Search(VariableName);
              if not Assigned(UseVariable) then begin
                Identifiers.Add(VariableName);
                while NextToken = ',' do begin
                  Expect(',');
                  Identifiers.Add(ReadToken);
                end;
              end else
                if not ValidateParams then
                  ErrorMessage(SRedefineIdentifier, [VariableName], -1);
              Expect(':');
              TypeName := ReadToken;
              UseType := Parent.Types.Search(TypeName);
              if not Assigned(UseType) then
                ErrorMessage(SUndefinedType, [TypeName], -1)
              else
                if ValidateParams then begin
                  for I := 0 to Identifiers.Count - 1 do begin
                    UseVariable := Parameters.Search(Identifiers[I]);
                    if Assigned(UseVariable) then
                      if UseVariable.ValueType <> UseType then ;
                        ErrorMessage(SParamDiffers, [Identifiers[I]]);
                  end;
                end else begin
                  for I := 0 to Identifiers.Count - 1 do
                    with TParameter(Items[Add(TParameter.Create)]) do
                    begin
                      Name := Identifiers[I];
                     ValueType := UseType;
                    end;

                end;
            end;
          end;
          if NextToken = ';' then Expect(';');
        end;
        Expect(')');
  finally
    Identifiers.Free;
  end;
end;

{ TParserIfThenElse }

function TPascalParser.ParseIfThenElse(var IfThenElse: TIfThenElse; SourceCode: TCommonBlock): Boolean;
begin
  if NextToken = 'if' then begin
    IfThenElse := TIfThenElse.Create;
    IfThenElse.CommonBlock := SourceCode;
    with IfThenElse do begin
      Expect('if');
      Condition.CommonBlock := CommonBlock;
      ParseExpression(Condition);
      Expect('then');
      Command := ParseCommand(CommonBlock);
      if NextToken = 'else' then
      begin
        Expect('else');
        ElseCommand := ParseCommand(CommonBlock);
      end;
    end;
    Result := True;
  end else Result := False;
end;

function TPascalParser.ParseForToDo(var ForToDo: TForToDo; SourceCode: TCommonBlock): Boolean;
var
  VariableName: string;
begin
  if NextToken = 'for' then begin
    ForToDo := TForToDo.Create;
    ForToDo.CommonBlock := SourceCode;
    with ForToDo do begin
      Expect('for');
      VariableName := ReadToken;
      ControlVariable := ForToDo.CommonBlock.Variables.Search(VariableName);
      if not Assigned(ControlVariable) then
        ErrorMessage(SUndefinedVariable, [VariableName], -1);
      Expect(':=');
      Start.CommonBlock := CommonBlock;
      ParseExpression(Start);
      Expect('to');
      Stop.CommonBlock := CommonBlock;
      ParseExpression(Stop);
      Expect('do');
      Command := ParseCommand(CommonBlock);
    end;
    Result := True;
  end else Result := False;
end;

function TPascalParser.ParseAssigment(var Assignment: TAssignment;
  SourceCode: TCommonBlock): Boolean;
var
  Variable: TVariable;
  IdentName: string;
begin
  if IsIdentificator(NextToken) then begin
    Variable := SourceCode.Variables.Search(NextToken);
    if Assigned(Variable) then begin
      // Variable assignment
      Assignment := TAssignment.Create;
      Assignment.CommonBlock := SourceCode;
      IdentName := ReadToken;
      Assignment.Target := SourceCode.Variables.Search(IdentName);
      Expect(':=');
      Assignment.Source := TExpression.Create;
      Assignment.Source.CommonBlock := SourceCode;
      ParseExpression(Assignment.Source);
      Result := True;
    end else Result := False;
  end else Result := False;
end;

function TPascalParser.ParseFunctionCall(var Call: TFunctionCall;
  SourceCode: TCommonBlock): Boolean;
var
  FunctionName: string;
begin
  if IsIdentificator(NextToken) then begin
    if Assigned(SourceCode.Functions.Search(NextToken)) then begin
      // Function call
      FunctionName := ReadToken;
      Call := TFunctionCall.Create;
      Call.CommonBlock := SourceCode;
      Call.FunctionRef := SourceCode.Functions.Search(FunctionName);
      if NextToken = '(' then begin
        Expect('(');
        with Call do begin
          ParameterExpression.Add(TExpression.Create);
          ParameterExpression.Last.CommonBlock := SourceCode;
          ParseExpression(ParameterExpression.Last);
        end;
        Expect(')');
      end;
      Result := True;
    end else Result := False;
  end else Result := False;
end;

{ TParserVariableList }

function TPascalParser.ParseVariableList(SourceCode: TVariableList; Exported: Boolean = False): Boolean;
var
  NewValueType: TType;
  TypeName: string;
  I: integer;
begin
  if NextToken = 'var' then begin
    Expect('var');
    with SourceCode do begin
      while IsIdentificator(NextToken) and (NextTokenType <> ttEndOfFile) do begin
        ParseVariable(SourceCode, Exported);
      end;
    end;
    Result := True;
  end else Result := False;
end;

{ TParserVariable }

procedure TPascalParser.ParseVariable(SourceCode: TVariableList; Exported: Boolean = False);
var
  VariableName: string;
  Variable: TVariable;
  TypeName: string;
  NewValueType: TType;
  Identifiers: TStringList;
  I: Integer;
begin
  try
    Identifiers := TStringList.Create;
  with SourceCode do begin
      Identifiers.Clear;
      VariableName := ReadToken;
      Variable := Search(VariableName);
      if not Assigned(Variable) then begin
        Identifiers.Add(VariableName);
        while NextToken = ',' do begin
          Expect(',');
          Identifiers.Add(ReadToken);
        end;
      end else
        ErrorMessage(SRedefineIdentifier, [VariableName], -1);
      Expect(':');
      TypeName := ReadToken;
      NewValueType := Parent.Types.Search(TypeName);
      if NewValueType = nil then
        ErrorMessage(SUndefinedType, [TypeName], -1)
      else
        for I := 0 to Identifiers.Count - 1 do
          with TVariable(Items[Add(TVariable.Create)]) do begin
            Name := Identifiers[I];
            ValueType := NewValueType;
          end;
      Expect(';');
  end;
  finally
    Identifiers.Free;
  end;
end;

{ TParserConstantList }

function TPascalParser.ParseConstantList(SourceCode: TConstantList; Exported: Boolean = False): Boolean;
begin
  if NextToken = 'const' then begin
    Expect('const');
    with SourceCode do begin
      while IsIdentificator(NextToken) do begin
        ParseConstant(SourceCode, Exported);
      end;
    end;
    Result := True;
  end else Result := False;
end;

function TPascalParser.ParseConstant(SourceCode: TConstantList;
  Exported: Boolean): Boolean;
var
  Identifiers: TStringList;
  NewValueType: TType;
  TypeName: string;
  ConstantName: string;
  Constant: TConstant;
  I: integer;
  ConstantValue: string;
begin
  with SourceCode do
  try
    Identifiers := TStringList.Create;
      ConstantName := ReadToken;
      Constant := Search(ConstantName);
      if not Assigned(Constant) then begin
        Identifiers.Add(ConstantName);
        while NextToken = ',' do begin
          Expect(',');
          Identifiers.Add(ReadToken);
        end;
      end else
        ErrorMessage(SRedefineIdentifier, [ConstantName], -1);
      if NextToken = ':' then begin
        Expect(':');
        TypeName := ReadToken;
        NewValueType := Parent.Types.Search(TypeName);
      end;
      Expect('=');
      ConstantValue := ReadToken;
      Expect(';');

      if NewValueType = nil then
        ErrorMessage(SUndefinedType, [TypeName], -1)
      else
        for I := 0 to Identifiers.Count - 1 do
          with TConstant(Items[Add(TConstant.Create)]) do
          begin
            Name := Identifiers[I];
            ValueType := NewValueType;
            Value := ConstantValue;
          end;
  finally
    Identifiers.Free;
  end;
end;

{ TParserTypeList }

function TPascalParser.ParseTypeList(SourceCode: TTypeList;
  Exported: Boolean = False; AssignSymbol: string = '='): Boolean;
var
  NewType: TType;
begin
  if NextToken = 'type' then begin
    Expect('type');
    with SourceCode do begin
      while IsIdentificator(NextToken) do begin
        NewType := ParseType(SourceCode, True, AssignSymbol);
        if Assigned(NewType) then begin
          NewType.Parent := SourceCode;
          Add(NewType);
        end;
        Expect(';');
      end;
    end;
    Result := True;
  end else Result := False
end;

{ TParserType }

function TPascalParser.ParseType(TypeList: TTypeList; ExpectName: Boolean = True;
  AssignSymbol: string = '='; ForwardDeclaration: Boolean = False): TType;
var
  Name: string;
  TypeName: string;
begin
  //with SourceCode do
  begin
    if ExpectName then begin
      Name := ReadToken;
      Expect(AssignSymbol);
    end;
    Result := ParseTypeEnumeration(TypeList, Name);
    if not Assigned(Result) then Result := ParseTypeRecord(TypeList, Name);
    if not Assigned(Result) then Result := ParseTypeClass(TypeList, Name);
    if not Assigned(Result) then Result := ParseTypeArray(TypeList, Name);
    if not Assigned(Result) then Result := ParseTypePointer(TypeList, Name);
    if not Assigned(Result) then Result := ParseTypeBase(TypeList, Name);
    if not Assigned(Result) then Result := ParseTypeSubType(TypeList, Name,
      ExpectName, ForwardDeclaration);
    if not Assigned(Result) then Result := ParseTypeSubRange(TypeList, Name);
    if not Assigned(Result) then begin
      ErrorMessage(SInvalidConstruction, []);
    end;
  end;
end;

function TPascalParser.ParseTypeSubType(TypeList: TTypeList; Name: string;
  ExpectName: Boolean; ForwardDeclaration: Boolean): TType;
var
  TypeName: string;
begin
  // Use existed type
  if NextTokenType = ttIdentifier then begin
    TypeName := ReadToken;
    if ExpectName then begin
      Result := TType.Create;
      TType(Result).Parent := TypeList;
      TType(Result).Name := Name;
      TType(Result).UsedType := TypeList.Search(TypeName);
      if not Assigned(TType(Result).UsedType) then
        ErrorMessage(SUndefinedType, [TypeName], -1);
    end else begin
      TType(Result) := TypeList.Search(TypeName);
      if not Assigned(TType(Result)) then begin
        if ForwardDeclaration then begin
          // ForwardDeclaration
          Result := TType.Create;
          TType(Result).Parent := TypeList;
          TType(Result).Name := TypeName;
          TType(Result).UsedType := nil;
        end else
          ErrorMessage(SUndefinedType, [TypeName], -1);
      end;
    end;
  end else Result := nil;
end;

function TPascalParser.ParseTypeBase(TypeList: TTypeList; Name: string): TType;
begin
      // Buildin base type construction
      if NextToken = 'type' then begin
        Expect('type');
        Result := TTypeInherited.Create;
        TTypeInherited(Result).Parent := TypeList;
        TTypeInherited(Result).Name := Name;
        if NextToken = '(' then begin
          Expect('(');
          TTypeInherited(Result).UsedType := ParseType(TypeList, False);
          Expect(')');
        end else TTypeInherited(Result).UsedType := nil;
      end else Result := nil;
end;

function TPascalParser.ParseTypePointer(TypeList: TTypeList; Name: string
  ): TType;
begin
  if NextToken = '^' then begin
    Expect('^');
    Result := TTypePointer.Create;
    TTypePointer(Result).Parent := TypeList;
    TTypePointer(Result).Name := Name;
    TTypePointer(Result).UsedType := ParseType(TypeList, False, '=', True);
  end else Result := nil;
end;

function TPascalParser.ParseTypeEnumeration(TypeList: TTypeList; Name: string): TType;
begin
  if NextToken = '(' then begin
      Expect('(');
      Result := TTypeEnumeration.Create;
      TTypeEnumeration(Result).Parent := TypeList;
      TTypeEnumeration(Result).Name := Name;
      with TTypeEnumeration(Result) do
      with TEnumItem(Items[Items.Add(TEnumItem.Create)]) do begin
        Name := ReadToken;
        if (NextToken = '=') and (NextTokenType = ttConstantNumber) then begin
          Expect('=');
          Index := StrToInt(ReadToken);
        end;
      end;
      while (NextToken = ',') and (NextTokenType <> ttEndOfFile) do
      begin
        Expect(',');
        with TTypeEnumeration(Result) do
        with TEnumItem(Items[Items.Add(TEnumItem.Create)]) do begin
          Name := ReadToken;
          if (NextToken = '=') and (NextTokenType = ttConstantNumber) then begin
            Expect('=');
            Index := StrToInt(ReadToken);
          end;
        end;
      end;
      Expect(')');
  end else Result := nil;
end;

function TPascalParser.ParseTypeRecord(TypeList: TTypeList; Name: string
  ): TType;
type
  TSectionType = (stVar, stType, stConst);
var
  Visibility: TTypeVisibility;
  SectionType: TSectionType;
begin
  if NextToken = 'record' then begin
    Expect('record');
    SectionType := stVar;
    Visibility := tvPublic;
    Result := TTypeRecord.Create;
    TTypeRecord(Result).Parent := TypeList;
    TTypeRecord(Result).CommonBlock.Parent := TypeList.Parent;
    TType(Result).Name := Name;
    while (NextToken <> 'end') and (NextTokenType <> ttEndOfFile) do begin
      if NextToken = 'public' then begin
        Expect('public');
        Visibility := tvPublic;
      end else
      if NextToken = 'private' then begin
        Expect('private');
        Visibility := tvPrivate;
      end else
      if NextToken = 'published' then begin
        Expect('published');
        Visibility := tvPublished;
      end else
      if NextToken = 'protected' then begin
        Expect('protected');
        Visibility := tvProtected;
      end else
      if NextToken = 'var' then begin
        SectionType := stVar;
        ParseVariableList(TTypeRecord(Result).CommonBlock.Variables, True);
      end else
      if NextToken = 'const' then begin
        SectionType := stConst;
        ParseConstantList(TTypeRecord(Result).CommonBlock.Constants, True)
      end else
      if NextToken = 'type' then begin
        SectionType := stType;
        ParseTypeList(TTypeRecord(Result).CommonBlock.Types, True, '=');
      end else
      if NextToken = 'procedure' then
        ParseFunctionList(TTypeRecord(Result).CommonBlock.Functions, True)
      else if NextToken = 'function' then
        ParseFunctionList(TTypeRecord(Result).CommonBlock.Functions, True)
      else begin
        if SectionType = stVar then begin
          if IsIdentificator(NextToken) then
            ParseVariable(TTypeRecord(Result).CommonBlock.Variables, True)
              else ReadToken;
            //TTypeRecord(Result).CommonBlock.Types.Add(ParseType(TypeList, True, ':'));
            //TType(TTypeRecord(Result).CommonBlock.Types.Last).Visibility := Visibility;
        end
        else if SectionType = stConst then
          ParseConstant(TTypeRecord(Result).CommonBlock.Constants, True)
        else if SectionType = stType then
          ParseType(TTypeRecord(Result).CommonBlock.Types, True, '=');
      end;
    end;
    Expect('end');
  end else Result := nil;
end;

function TPascalParser.ParseTypeClass(TypeList: TTypeList; Name: string
  ): TType;
begin
  if NextToken = 'class' then begin
      Expect('class');
      Result := TTypeClass.Create;
      TTypeClass(Result).Parent := TypeList;
      TTypeClass(Result).Name := Name;
      if NextToken <> ';' then begin
        while (NextToken <> 'end') and (NextTokenType <> ttEndOfFile) do
        begin
          TTypeClass(Result).Items.Add(ParseType(TypeList, True, ':'));
          Expect(';');
        end;
        Expect('end');
      end;
  end else Result := nil;
end;

function TPascalParser.ParseTypeArray(TypeList: TTypeList; Name: string
  ): TType;
var
  UseName: string;
  UseType: TType;
begin
  if NextToken = 'array' then begin
    Expect('array');
    Result := TTypeArray.Create;
    TTypeArray(Result).Parent := TypeList;
    TType(Result).Name := Name;
    if NextToken = '[' then begin
      Expect('[');
      UseName := NextToken;
      if NextTokenType = ttIdentifier then begin
        UseType := TypeList.Parent.Types.Search(UseName);
        if not Assigned(TTypeArray(Result).IndexType) then
          ErrorMessage(SUndefinedType, [UseName], -1) else
          TTypeArray(Result).IndexType := UseType;
      end else
      if NextTokenType = ttConstantNumber then begin
        UseType := ParseTypeSubRange(TypeList, Name);
        if not Assigned(UseType) then begin
          ErrorMessage(SInvalidConstruction, [], -1);
        end;
      end;
      Expect(']');
    end;
    Expect('of');
    UseName := NextToken;
    TTypeArray(Result).ItemType := ParseType(TypeList, False);
    if not Assigned(TTypeArray(Result).ItemType) then
      ErrorMessage(SUndefinedType, [UseName], -1);
  end else Result := nil;
end;

function TPascalParser.ParseTypeSubRange(TypeList: TTypeList; Name: string
  ): TType;
var
  UseName: string;
begin
  if NextTokenType = ttConstantString then begin
    Result := TTypeSubRange.Create;
    TTypeSubRange(Result).Bottom := ReadToken;
    Expect('..');
    TTypeSubRange(Result).Top := ReadToken;
  end else
  if NextTokenType = ttConstantNumber then begin
    Result := TTypeSubRange.Create;
    TTypeSubRange(Result).Bottom := ReadToken;
    Expect('..');
    TTypeSubRange(Result).Top := ReadToken;
  end else Result := nil;
end;

constructor TPascalParser.Create;
begin
  inherited;
end;

destructor TPascalParser.Destroy;
begin
  inherited Destroy;
end;

{ TParserUsedModuleList }

function TPascalParser.ParseUses(SourceCode: TUsedModuleList; AExported: Boolean = False): Boolean;
var
  NewUsedModule: TUsedModule;
begin
  if NextToken = 'uses' then begin
    Expect('uses');
    with TUsedModule(SourceCode.Items[SourceCode.Add(TUsedModule.Create)]) do begin
      Name := ReadToken;
      if NextToken = 'in' then begin
        Expect('in');
        Location := ReadToken;
      end else Location := Name + '.pas';
      Module := SourceCode.ParentModule.ParentProgram.Modules.Search(Name);
      if not Assigned(Module) then begin
        if ParseFile(Name) then begin
          Module := SourceCode.ParentModule.ParentProgram.Modules.Search(Name);
          Exported := AExported;
        end else begin
          ErrorMessage(SUnitNotFound, [Name], -2);
          SourceCode.Delete(SourceCode.Count - 1);
        end;
      end;
    end;
    while NextToken = ',' do begin
      Expect(',');
      with TUsedModule(SourceCode.Items[SourceCode.Add(TUsedModule.Create)]) do
      begin
        Name := ReadToken;
        if NextToken = 'in' then begin
          Expect('in');
          Location := ReadToken;
        end else Location := Name + '.pas';
        Module := SourceCode.ParentModule.ParentProgram.Modules.Search(Name);
        if not Assigned(Module) then begin
          if not ParseFile(Name) then begin
            ErrorMessage(SUnitNotFound, [Name], -2);
            SourceCode.Delete(SourceCode.Count - 1);
          end;
        end;
      end;
    end;
    Expect(';');
    Result := True;
  end else Result := False;
end;


end.

