unit UCompiler;

{$MODE Delphi}

interface

uses
  SysUtils, Variants, Classes,
  Dialogs, USourceCode, UProducer, UPascalParser, UParser,
  UProducerDynamicC, Contnrs, UProducerTreeView, UProducerASM8051,
  UProducerPascal, UProducerGCCC;

type
  TProducerType = (ptGCCC, ptDynamicC, ptPascal, ptAssembler, ptXML);
  TAnalyzerType = (ptCustomPascal, ptDelphi);

  TErrorMessage = class
    Text: string;
    Position: TPoint;
    FileName: string;
  end;

  TCompilerTarget = class
    Vendor: string;
    Model: string;
    Architecture: string;
  end;

  { TCompilerList }

  TCompilerTargetList = class(TObjectList)
    procedure Add(Vendor, Model, Architecture: string);
  end;

  TCompiler = class
  private
    FOnErrorMessage: TErrorMessageEvent;
    FProducerType: TProducerType;
    procedure ErrorMessage(Text: string; Position: TPoint; FileName: string);
    procedure SetProducerType(const AValue: TProducerType);
  public
    ProgramCode: TProgram;
    Parser: TPascalParser;
    Producer: TProducer;
    ErrorMessages: TObjectList;
    CompiledFolder: string;
    SupportedTargets: TCompilerTargetList;
    Target: TCompilerTarget;
    TargetFolder: string;
    constructor Create;
    destructor Destroy; override;
    procedure Init;
    procedure Compile(ModuleName: string; Source: TStringList);
    property OnErrorMessage: TErrorMessageEvent read FOnErrorMessage
      write FOnErrorMessage;
    property ProducerType: TProducerType read FProducerType
      write SetProducerType;
  end;

const
  ProducerTypeName: array[TProducerType] of string = (
    'GCC C', 'Rabbit Dynamic C', 'Generic Pascal', 'Assembler', 'XML');
  AnalyzerTypeName: array[TAnalyzerType] of string = (
    'Custom pascal', 'Delphi');

implementation

{ TCompiler }

procedure TCompiler.Compile(ModuleName: string; Source: TStringList);
var
  NewModule: TModule;
  ProducedCode: TStringList;
  I: Integer;
  TargetFileName: string;
begin
  try
    ProducedCode := TStringList.Create;
    Parser.FileName := ModuleName;
    Parser.SourceCodeText := Source;
    Parser.Process;
    //ShowMessage(IntToHex(Integer(Addr(Parser.OnGetSource)), 8));
    Parser.ParseModule(ProgramCode);
    with ProgramCode do
    for I := 0 to Modules.Count - 1 do begin
      Producer.Produce(TModule(Modules[I]));
      Producer.AssignToStringList(ProducedCode);
      TargetFileName := TargetFolder + DirectorySeparator +
        CompiledFolder + DirectorySeparator + Producer.ClassName +
        DirectorySeparator + TModule(Modules[I]).TargetFile;
      ForceDirectories(ExtractFileDir(TargetFileName));
      ProducedCode.SaveToFile(TargetFileName);
    end;
  finally
    ProducedCode.Free;
  end;
end;

constructor TCompiler.Create;
begin
  SupportedTargets := TCompilerTargetList.Create;
  SupportedTargets.Add('AMD', 'Athlon II', 'amd64');
  SupportedTargets.Add('Zilog', 'Z80', 'Z80');
  SupportedTargets.Add('Rabbit', 'RC3000', 'Z180');
  SupportedTargets.Add('Atmel', 'ATmega8', 'AVR');
  SupportedTargets.Add('Atmel', 'UC3B0256', 'AVR32');
  SupportedTargets.Add('Dallas', 'DS89C450', '8052');

  ProgramCode := TProgram.Create;
  Producer := TProducerGCCC.Create;
  Parser := TPascalParser.Create;
  Parser.OnErrorMessage := ErrorMessage;
  ErrorMessages := TObjectList.Create;
  CompiledFolder := 'Compiled';
end;

procedure TCompiler.Init;
begin
  ErrorMessages.Clear;
  ProgramCode.Modules.Clear;
end;

destructor TCompiler.Destroy;
begin
  ProgramCode.Free;
  Parser.Free;
  Producer.Free;
  ErrorMessages.Free;
  SupportedTargets.Free;
end;

procedure TCompiler.ErrorMessage(Text: string; Position: TPoint; FileName: string);
var
  NewMessage: TErrorMessage;
begin
  NewMessage := TErrorMessage.Create;
  NewMessage.Text := Text;
  NewMessage.Position := Position;
  NewMessage.FileName := FileName;
  ErrorMessages.Add(NewMessage);
  if Assigned(FOnErrorMessage) then FOnErrorMessage(Text, Position, FileName);
end;

procedure TCompiler.SetProducerType(const AValue: TProducerType);
begin
  if FProducerType = AValue then Exit;
  FProducerType := AValue;
  Producer.Free;
  case AValue of
    ptGCCC: Producer := TProducerGCCC.Create;
    ptDynamicC: Producer := TProducerDynamicC.Create;
    ptPascal: Producer := TProducerPascal.Create;
    ptAssembler: Producer := TProducerGCCC.Create;
    ptXML: Producer := TProducerTreeView.Create;
  end;
end;

{ TCompilerTargetList }

procedure TCompilerTargetList.Add(Vendor, Model, Architecture: string);
var
  NewItem: TCompilerTarget;
begin
  NewItem := TCompilerTarget.Create;
  NewItem.Vendor := Vendor;
  NewItem.Model := Model;
  NewItem.Architecture := Architecture;
  inherited Add(NewItem);
end;

end.
