unit UProject;

{$mode objfpc}{$H+}
{$MACRO ON}

interface

uses
  Classes, SysUtils, Dialogs, DOM, XMLWrite, XMLRead, FileUtil;

type

  { TProjectNode }

  TProjectNode = class
    Parent: TProjectNode;
    Name: string;
    ProducedFileName: string;
    procedure SaveToXMLNode(Node: TDOMNode); virtual;
    procedure LoadFromXMLNode(Node: TDOMNode); virtual;
    procedure Load; virtual;
    procedure Save; virtual;
    function GetDir(IncludeRoot: Boolean = False): string; virtual;
  end;


  // TListProjectNode = TGObjectList<Integer, TProjectNode>
  {$DEFINE TGObjectListIndex := Integer}
  {$DEFINE TGObjectListItem := TProjectNode}
  {$DEFINE TGObjectListList := TObjectListProjectNode}
  {$DEFINE TGObjectList := TListProjectNode}
  {$DEFINE TGObjectListSortCompare := TObjectListSortCompareProjectNode}
  {$DEFINE TGObjectListStringConverter := TObjectListStringConverterProjectNode}
  {$DEFINE INTERFACE}
  {$I 'GenericObjectList.inc'}

  { TProjectFile }

  TProjectFile = class(TProjectNode)
    Source: TStringList;
    constructor Create;
    destructor Destroy; override;
    procedure SaveToXMLNode(Node: TDOMNode); override;
    procedure LoadFromXMLNode(Node: TDOMNode); override;
    procedure Load; override;
    procedure Save; override;
  end;

  { TProjectGroup }

  TProjectGroup = class(TProjectNode)
    Items: TListProjectNode;
    constructor Create;
    destructor Destroy; override;
    procedure Load; override;
    procedure Save; override;
    procedure SaveToXMLNode(Node: TDOMNode); override;
    procedure LoadFromXMLNode(Node: TDOMNode); override;
    function SearchFile(FileName: string): TProjectFile;
  end;

  TProject = class(TProjectGroup)
    RootDir: string;
    procedure LoadFromFile(FileName: string);
    procedure SaveToFile(FileName: string);
    function GetDir(IncludeRoot: Boolean = False): string; override;
    procedure Load; override;
    procedure Save; override;
    procedure SaveToXMLNode(Node: TDOMNode); override;
    procedure LoadFromXMLNode(Node: TDOMNode); override;
    procedure Clear;
  end;

implementation


{$DEFINE IMPLEMENTATION_USES}
{$INCLUDE 'GenericObjectList.inc'}

// TListProjectNode = TGObjectList<Integer, TProjectNode>
{$DEFINE TGObjectListIndex := Integer}
{$DEFINE TGObjectListItem := TProjectNode}
{$DEFINE TGObjectListList := TObjectListProjectNode}
{$DEFINE TGObjectList := TListProjectNode}
{$DEFINE TGObjectListSortCompare := TObjectListSortCompareProjectNode}
{$DEFINE TGObjectListStringConverter := TObjectListStringConverterProjectNode}
{$DEFINE IMPLEMENTATION}
{$INCLUDE 'GenericObjectList.inc'}

{ TProjectGroup }

constructor TProjectGroup.Create;
begin
  inherited;
  Items := TListProjectNode.Create;
end;

destructor TProjectGroup.Destroy;
begin
  Items.Free;
  inherited Destroy;
end;

procedure TProjectGroup.Load;
var
  I: Integer;
begin
  inherited Load;
  for I := 0 to Items.Count - 1 do
    TProjectNode(Items[I]).Load;
end;

procedure TProjectGroup.Save;
var
  I: Integer;
begin
  for I := 0 to Items.Count - 1 do
    TProjectNode(Items[I]).Save;
end;

procedure TProjectGroup.SaveToXMLNode(Node: TDOMNode);
begin
  inherited SaveToXMLNode(Node);
end;

function TProjectGroup.SearchFile(FileName: string): TProjectFile;
var
  NamePart: string;
  I: Integer;
begin
  if Pos('/', FileName) > 0 then
    NamePart := Copy(FileName, 1, Pos('/', FileName))
    else NamePart := FileName;
  I := 0;
  while (I < Items.Count) and (ExtractFileNameOnly(TProjectNode(Items[I]).Name) <> NamePart) do Inc(I);
  if I < Items.Count then begin
    if TProjectNode(Items[I]) is TProjectFile then
      Result := TProjectFile(Items[I])
    else begin
      Delete(FileName, 1, Length(FileName) + 1);
      TProjectGroup(Items[I]).SearchFile(FileName);
    end;
  end else Result := nil;
end;

{ TProjectFile }

constructor TProjectFile.Create;
begin
  Source := TStringList.Create;
end;

destructor TProjectFile.Destroy;
begin
  Source.Free;
  inherited Destroy;
end;

procedure TProjectFile.SaveToXMLNode(Node: TDOMNode);
begin
  inherited SaveToXMLNode(Node);
end;

procedure TProjectFile.LoadFromXMLNode(Node: TDOMNode);
begin
  inherited LoadFromXMLNode(Node);
  Load;
end;

procedure TProjectFile.Load;
begin
  inherited Load;
  Source.LoadFromFile(Parent.GetDir(True) + Name);
end;

procedure TProjectFile.Save;
begin
  inherited Save;
  Source.SaveToFile(Parent.GetDir(True) + Name);
end;

{ TProjectNode }

procedure TProjectNode.SaveToXMLNode(Node: TDOMNode);
begin

end;

procedure TProjectNode.LoadFromXMLNode(Node: TDOMNode);
begin
  Name := UTF8Encode(string(TDOMElement(Node).GetAttribute('Name')));
end;

procedure TProjectNode.Load;
begin

end;

procedure TProjectNode.Save;
begin

end;

function TProjectNode.GetDir(IncludeRoot: Boolean = False): string;
begin
  if Assigned(Parent) then Result := Parent.GetDir(IncludeRoot) + Name + '/'
  else Result := Name + '/';
end;

{ TProject }

procedure TProject.LoadFromFile(FileName: string);
var
  Doc: TXMLDocument;
  NewNode: TDOMNode;
  NewSubNode: TDOMNode;
  I: Integer;
  Text: string;
begin
  try
    Clear;
    RootDir := ExtractFileDir(FileName) + '/';
    Name := ExtractFileName(FileName);
    Name := Copy(Name, 1, Length(Name) - Length(ExtractFileExt(FileName)));

    ReadXMLFile(Doc, UTF8Decode(FileName));

    NewNode := Doc.DocumentElement.FindNode('Files');
    if Assigned(NewNode) then
      LoadFromXMLNode(NewNode);
  finally
    Doc.Free;
  end;
end;

procedure TProjectGroup.LoadFromXMLNode(Node: TDOMNode);
var
  Child: TDOMNode;
  ProjectNode: TProjectNode;
  NodeType: string;
begin
  Name := UTF8Encode(TDOMElement(Node).GetAttribute('Name'));
  Child := Node.FirstChild;
  while Assigned(Child) do begin
    NodeType := TDOMElement(Child).GetAttribute('Type');
    if NodeType = 'Folder' then
      ProjectNode := TProjectGroup.Create
    else if NodeType = 'File' then
      ProjectNode := TProjectFile.Create
    else ProjectNode := nil;
    if Assigned(ProjectNode) then begin
      ProjectNode.Parent := Self;
      ProjectNode.LoadFromXMLNode(Child);
      Items.Add(ProjectNode);
    end;

    Child := Child.NextSibling;
  end;
end;

procedure TProject.SaveToFile(FileName: string);
begin

end;

function TProject.GetDir(IncludeRoot: Boolean): string;
begin
  if IncludeRoot then Result := RootDir else Result := '';
end;

procedure TProject.Load;
begin
  inherited Load;
end;

procedure TProject.Save;
begin
  inherited Save;
end;

procedure TProject.SaveToXMLNode(Node: TDOMNode);
begin
  inherited SaveToXMLNode(Node);
end;

procedure TProject.LoadFromXMLNode(Node: TDOMNode);
begin
  inherited LoadFromXMLNode(Node);
end;

procedure TProject.Clear;
begin
  Items.Clear;
end;

end.

